// Copyright 2017 The Cockroach Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
// implied. See the License for the specific language governing
// permissions and limitations under the License.

package storage

import (
	"fmt"
	"strings"
	"testing"

	"golang.org/x/net/context"

	"github.com/cockroachdb/cockroach/pkg/roachpb"
	"github.com/cockroachdb/cockroach/pkg/storage/engine"
	"github.com/cockroachdb/cockroach/pkg/storage/engine/enginepb"
	"github.com/cockroachdb/cockroach/pkg/util/leaktest"
	"github.com/cockroachdb/cockroach/pkg/util/uuid"
)

func TestDeclareKeysResolveIntent(t *testing.T) {
	defer leaktest.AfterTest(t)()

	const id = "f90b99de-6bd2-48a3-873c-12fdb9867a3c"
	txnMeta := enginepb.TxnMeta{}
	{
		var err error
		txnMeta.ID, err = uuid.FromString(id)
		if err != nil {
			t.Fatal(err)
		}
	}
	abortCacheKey := fmt.Sprintf(`1 1: /{Local/RangeID/99/r/AbortCache/"%s"-Min}`, id)
	desc := roachpb.RangeDescriptor{
		RangeID:  99,
		StartKey: roachpb.RKey("a"),
		EndKey:   roachpb.RKey("a"),
	}
	tests := []struct {
		status      roachpb.TransactionStatus
		poison      bool
		expDeclares bool
	}{
		{
			status:      roachpb.ABORTED,
			poison:      true,
			expDeclares: true,
		},
		{
			status:      roachpb.ABORTED,
			poison:      false,
			expDeclares: true,
		},
		{
			status:      roachpb.COMMITTED,
			poison:      true,
			expDeclares: false,
		},
		{
			status:      roachpb.COMMITTED,
			poison:      false,
			expDeclares: false,
		},
	}
	ctx := context.Background()
	engine := engine.NewInMem(roachpb.Attributes{}, 1<<20)
	defer engine.Close()
	for _, ranged := range []bool{false, true} {
		t.Run(fmt.Sprintf("ranged=%t", ranged), func(t *testing.T) {
			for _, test := range tests {
				t.Run("", func(t *testing.T) {
					ri := roachpb.ResolveIntentRequest{
						IntentTxn: txnMeta,
						Status:    test.status,
						Poison:    test.poison,
					}
					ri.Key = roachpb.Key("b")
					rir := roachpb.ResolveIntentRangeRequest(ri)
					rir.EndKey = roachpb.Key("c")

					ac := NewAbortCache(desc.RangeID)

					var spans SpanSet
					batch := engine.NewBatch()
					batch = makeSpanSetBatch(batch, &spans)
					defer batch.Close()

					var h roachpb.Header
					h.RangeID = desc.RangeID

					cArgs := CommandArgs{Header: h}
					cArgs.EvalCtx.repl = &Replica{abortCache: ac}

					if !ranged {
						cArgs.Args = &ri
						declareKeysResolveIntent(desc, h, &ri, &spans)
						if _, err := evalResolveIntent(ctx, batch, cArgs, &roachpb.ResolveIntentResponse{}); err != nil {
							t.Fatal(err)
						}
					} else {
						cArgs.Args = &rir
						declareKeysResolveIntentRange(desc, h, &rir, &spans)
						if _, err := evalResolveIntentRange(ctx, batch, cArgs, &roachpb.ResolveIntentRangeResponse{}); err != nil {
							t.Fatal(err)
						}
					}

					if s := spans.String(); strings.Contains(s, abortCacheKey) != test.expDeclares {
						t.Errorf("expected abort cache declared: %t, but got spans\n%s", test.expDeclares, s)
					}
				})
			}
		})
	}
}
